/*********************************************************************************
** This is a Motion tracking and visualizing example.
** This would be used for MAT265 course by Prof. George Legrady in 2016 Spring.
** 
** This program only runs in Processing 2 since some of libraries are not supported in other version of Processing. 
** Please sucessfully connect a Kinect or Xtion before run this program.
**
**
** Author: Hilda Chang HE
** Previous Version by: Donghao Ren, Rodger Jieliang Luo 
** Date: 2016 Winter: Jan - March
** Version: Concise version from Hilda
**
**
** Thanks for help from Rodger and Donghao, and guide by Prof. George Legrady.
**
*********************************************************************************/

import peasy.*;
import SimpleOpenNI.*;
import java.util.Iterator;

SimpleOpenNI context;

PeasyCam cam;

// width, height are given as window sizes
int screen_width = 1024; 
int screen_height = 700; 
float depth = 500;
float camDepth = 2000;

// Distance between the camera and the screen in the virtual space.
float camera_screen_distance = 4500; //500

//Boundary and initial values 
float left = -100;
float right = 100;
float top = -100;
float bottom = 100;
float back = 50;
float front = -500;

// Initialize OpenNI or not (for debugging).
boolean initialize_openni = true;
// Show hand positions (for debugging).
boolean show_Com_vectors = true;
boolean draw_boundary = true;
boolean draw_path = true;

// User tracking information.
HashMap<Integer, UserTrackingInfo> user_tracking_info = new HashMap<Integer, UserTrackingInfo>();

//avoid concurency modification
Object lock = new Object();

// Historical Users Information 
HistoricalUsers historyInfo = new HistoricalUsers();

// Previous update time.
double previous_update_time;

//User location
PVector com;

//initiate time
float time;
int fadeSpeed = 100; // larger -> faster
int alphaThreshold = 80;//transparency

//distance measurement
float distance = 0;


void setup() {
    // Initialize screen.
    // OpenGL rendering.
    size(screen_width, screen_height, OPENGL);
    
    // Initialize OpenNI.
    if(initialize_openni) {
        context = new SimpleOpenNI(this);
        if(context.isInit() == false) {
            println("Can't init SimpleOpenNI, maybe the camera is not connected!");
            exit();
            return;
        }
    } else {
        context = null;
    }

    if(context != null) {
        // Disable mirror.
        context.setMirror(true);

        // Seems it stops randomly crashing if we enable depth.
        context.enableDepth();

        // Enable user tracking.
        context.enableUser();
        
        context.update();
        
        com = new PVector(); // user location
    }
    //peasyCam
      cam = new PeasyCam(this, width / 2, height / 2, -depth - camera_screen_distance, back + camDepth);
    
    //set frameRate
    frameRate(20);
    
    //set time
    time = 0.00;
    
}

// Translate OpenNI space into image space.
// GL: translates the sensing distance from /real space sensing to virtual space
PVector translateOpenNIPosition(PVector position) {
    float x = -position.x;
    float y = -position.y;
    // Mirrored version (Z-direction).
    // float z = depth - position.z;
    // Non-Mirrored version.
    // distance range adjust here ( / 1 -> 2 meters, / 3 -> 4 meters).
     float z = position.z;// (position.z - 1000);
    //println(position.z);
    return new PVector(x, y, z);
}

// Update hand positions.
// We don't actually use this, but this function could be useful.
void updateHandPoses() {
    // Calculate dt (in seconds).
    double current_time = millis();
    float dt = (float)(current_time - previous_update_time) / 1000.0;
    previous_update_time = current_time;

    // Read OpenNI data and update user hand positions and velocities.
    if(context != null) {
        // Get forces.
        int[] user_list = context.getUsers();
        for(int i = 0; i < user_list.length; i++) {
            if(context.isTrackingSkeleton(user_list[i])) {
                int user_id = user_list[i];
                PVector left_hand_position = new PVector();
                PVector right_hand_position = new PVector();
                float left_hand_confidence = context.getJointPositionSkeleton(user_id, SimpleOpenNI.SKEL_LEFT_HAND, left_hand_position);
                float right_hand_confidence = context.getJointPositionSkeleton(user_id, SimpleOpenNI.SKEL_RIGHT_HAND, right_hand_position);
                left_hand_position = translateOpenNIPosition(left_hand_position);
                right_hand_position = translateOpenNIPosition(right_hand_position);
                UserTrackingInfo info = user_tracking_info.get(user_id);

            }
        }
    }
}

//Update user position 
void updateUserPoses(){
    if(context != null) {
          int[] user_list = context.getUsers();
          for(int i = 0; i < user_list.length; i++) {
                //println("test for distance use " + i +":  "+ user_list[i]);
             if(context.getCoM(user_list[i], com)){
                println("user com: " + com);
                int user_id = user_list[i];
                //println("Current :" + com);
                if(com != null && user_tracking_info.get(user_id) != null){
                    user_tracking_info.get(user_id).update(new PVector(com.x, com.y, com.z));
                }
            }
        }
    }  
}

//Update the box boundary, which is the border of our movement. 
void updateBoxBoundary(){
    // Read OpenNI data and update boundary
    if(context != null) {
        // iteration
        int[] user_list = context.getUsers();
        for(int i = 0; i < user_list.length; i++) {
            if(context.getCoM(user_list[i], com)){
                PVector p = translateOpenNIPosition(com);
                //println("User Location: x(" + p.x + ") y(" + p.y + ") z(" + p.z+")");
                if(p.x < left ) {left = p.x;println("Update LEFT, x = " + left );}
                else if(p.x > right) {right = p.x; println("Update RIGHT, x = ", + right );}
                if(p.y < top ) {top = p.y;println("Update TOP, y = " + top );}
                else if(p.y > bottom ) {bottom = p.y;println("Update BOTTOM, y = " + bottom );}
                if(p.z > back) {
                    back = p.z;println("Update BACK, z = " + back );
                    //change camera distance
                    cam.setDistance(back + camDepth);
                }
            }
        }
    }
}

  

void draw() {
  if(context != null) context.update();
    synchronized (lock) {
        background(50);
        
        //updateHandPoses();
        updateBoxBoundary();
        updateUserPoses();
    
        translate(width / 2, height / 2, -depth - camera_screen_distance);
        
        Iterator<Integer> it;
        //draw user position 
        if(show_Com_vectors) {
            stroke(255, 255, 255);
            fill(255, 255, 255);
            
            it = user_tracking_info.keySet().iterator();
            if(context != null) {
               // iteration
               int[] user_list = context.getUsers(); // =========================
               for(int i = 0; i < user_list.length; i++) {
                   if(context.getCoM(user_list[i], com)){
                      PVector p = translateOpenNIPosition(com);
                      //println("User Location: x(" + p.x + ") y(" + p.y + ") z(" + p.z+")");
                      drawBox(p.x, p.y, p.z);
                      float dist = sqrt((p.x/22.7)*(p.x/22.7) + (p.y/22.7)*(p.y/22.7) + (p.z/22.7)*(p.z/22.7));
                      fill(255);
                      textSize(24);
                      text("Distance: " + (int)dist + " inch", p.x-25, p.y+25, p.z); 
                   }
                }
            }
        }
        if(draw_path){
            it = user_tracking_info.keySet().iterator();
            if(context != null) {
                while(it.hasNext()) {
                    int user_id = it.next();
                    UserTrackingInfo info = user_tracking_info.get(user_id);
                    stroke(info.colour.x, info.colour.y, info.colour.z);
                    for(int i = 0; i < info.positions.size(); i++){
                        //println("i: " + i);
                        PVector p = translateOpenNIPosition(info.positions.get(i));
                        //println("path debug: x " +info.positions.get(i).x+ "  y " +info.positions.get(i).y+ "  z "+info.positions.get(i).z);
                        drawCross(p.x, p.y, p.z);   
                    }
                }
            }
    
            //draw history data
            if(historyInfo.counter > 0){
                //for(int i = 0; i < historyInfo.counter; i++){
                for(int i = historyInfo.counter - 1; i >= 0; i--){
                    float uTime = historyInfo.userTime.get(i);
                    float alpha = 255;
                    ArrayList<PVector> cUser = historyInfo.historyUsers.get(i);
                    //draw one user data
                    alpha = 255 - fadeSpeed * (time - uTime);
                    if(alpha < alphaThreshold){
                      alpha = alphaThreshold;
                    }
                    for(int j = 0; j< cUser.size(); j++){
                        drawCross(cUser.get(j), historyInfo.userColors.get(i), alpha);
                    }
                }
            }
        }
        
        if(draw_boundary){
          stroke(250,250,70);
          beginShape(LINES);
          //z-axis
          vertex(left, top, front);
          vertex(left, top, back);
          vertex(right, top, front);
          vertex(right, top, back);
          vertex(right, bottom, front);
          vertex(right, bottom, back);
          vertex(left, bottom, front);
          vertex(left, bottom, back); 
          //x-axis
          vertex(left, top, front);
          vertex(right, top, front);
          vertex(left, top, back);
          vertex(right, top, back);
          vertex(left, bottom, back);
          vertex(right, bottom, back);
          vertex(left, bottom, front);
          vertex(right, bottom, front);
          //y-axis
          vertex(left, top, front);
          vertex(left, bottom, front);
          vertex(right, top, front);
          vertex(right, bottom, front);
          vertex(left, top, back);
          vertex(left, bottom, back);
          vertex(right, top, back);
          vertex(right, bottom, back);
          endShape();
        }
    }
    
    time += 0.01;
}

//draw a small cude by given parameters
void drawBox(float x, float y, float z) {
    pushMatrix();
    translate(x, y, z);
    box(20, 20, 20);
    popMatrix();
}

//draw a small cross by given parameters
void drawCross(float x, float y, float z){
    beginShape(LINES);
    vertex(x - 50, y, z);
    vertex(x + 50, y, z);
    vertex(x, y-50, z);
    vertex(x, y+50, z);
    vertex(x, y, z-50);
    vertex(x, y, z+50);
    endShape();
}

//draw a small cross by given parameters 
void drawCross(PVector p, PVector colour, float alpha){
    stroke(colour.x, colour.y, colour.z, alpha);
    beginShape(LINES);
    vertex(p.x - 50, p.y, p.z);
    vertex(p.x + 50, p.y, p.z);
    vertex(p.x, p.y-50, p.z);
    vertex(p.x, p.y+50, p.z);
    vertex(p.x, p.y, p.z-50);
    vertex(p.x, p.y, p.z+50);
    endShape();
}

// -----------------------------------------------------------------
// SimpleOpenNI user events

//event handler when found new user
void onNewUser(SimpleOpenNI current_context, int user_id) {
    synchronized (lock) {
        println("New user: id = " + user_id + ", started tracking.");
        user_tracking_info.put(user_id, new UserTrackingInfo());
        //context.startTrackingSkeleton(user_id);
    }
}

//event handler when lost track of this user
void onLostUser(SimpleOpenNI current_context, int user_id) {
    synchronized (lock) {
        println("Lost user: id = " + user_id);
        UserTrackingInfo info = user_tracking_info.get(user_id);
        historyInfo.addUserData(info.positions, info.colour, time);
        user_tracking_info.remove(user_id);
    }
}
